/* 
Copyright (C) 2014 - 2017 Texas Instruments Incorporated - http://www.ti.com/

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the
distribution.

Neither the name of Texas Instruments Incorporated nor the names of
its contributors may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ======== peripherals.c ========
 *
 */
#include "peripherals.h"
#include "USB_API/USB_CDC_API/UsbCdc.h"
#include "USB_app/usbConstructs.h"

enum active_peripherals active_peripheral = None;

uint8_t rxBuffer[RX_BUFFER_SIZE]; 
uint8_t txBuffer[TX_BUFFER_SIZE];
uint8_t *txBuffer_read_ptr = txBuffer;
uint8_t *txBuffer_write_ptr = txBuffer;

uint8_t  UARTtoCDC_CircularBuffer[BSL_BUFFER_SIZE];
uint16_t UARTtoCDC_RXPtr = 0;
uint16_t UARTtoCDC_TXPtr = 0;

uint32_t current_baudrate = 0;
uint8_t isE4xx = 0;

/*** LEDs function during startup *********************************************/

void startup_led_sequence()
{
  uint8_t led_blink_count;
  for(led_blink_count = LED_BLINK_COUNT; led_blink_count > 0; --led_blink_count)
  {
    LED1_ON
    LED2_ON
    __delay_cycles(LED_BLINK_DELAY);
    LED1_OFF
    LED2_OFF
    __delay_cycles(LED_BLINK_DELAY);
  }
}

/*** BSL entry sequence *******************************************************/

void BSL_invoke_sequence()
{
  /* BSL invoke sequence 
   *
   *       H -----------+     +--+  +-----+
   * TEST  L            +-----+  +--+     +--------
   *
   *       H --------------+           +-----------
   * RST   L               +-----------+
   *
   *
   *       H -----------------+  +--+     +--------
   * TCK   L                  +--+  +-----+
   */

  // Set RST, TST and TCK pin to output
  ENTRY_SEQ_PDIR |= (RESET_PIN + TEST_PIN + TCK_PIN);

  // (0)Start with RST high, TEST high, TCK high
  ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN + TCK_PIN;
  __delay_cycles(INVOKE_DELAY_100US);
  __delay_cycles(INVOKE_DELAY_100US);

  // (1)RST high, TEST low, TCK high
  ENTRY_SEQ_POUT = RESET_PIN + TCK_PIN;
  __delay_cycles(INVOKE_DELAY_100US);

  // (2)RST low, TEST low, TCK high
  ENTRY_SEQ_POUT = TCK_PIN;
  __delay_cycles(INVOKE_DELAY_2US);

  // (3)RST low, TEST high, TCK low
  ENTRY_SEQ_POUT = TEST_PIN ;
  __delay_cycles(INVOKE_DELAY_110US);

  // (4)RST low, TEST low, TCK high
  ENTRY_SEQ_POUT = TCK_PIN;
  __delay_cycles(INVOKE_DELAY_10US);

  // (5)RST low, TEST high, TCK low
  ENTRY_SEQ_POUT =TEST_PIN;
  __delay_cycles(INVOKE_DELAY_100US);

  // (6)RST high, TEST high, TCK low
  ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN;
  __delay_cycles(INVOKE_DELAY_100US);
  
  // (7)RST high, TEST low, TCK high
  ENTRY_SEQ_POUT = RESET_PIN + TCK_PIN;
  __delay_cycles(INVOKE_DELAY_100US);
  __delay_cycles(INVOKE_DELAY_100US);

  // Set RST, TST and TCK pin to input
  ENTRY_SEQ_PDIR &= ~(RESET_PIN + TEST_PIN + TCK_PIN);
}

/*** Microsecond delay function for Legacy BSL ******************************************************/
void BSL_delay_ms(uint8_t time)
{
  uint8_t i;
  
  for(i = 0; i<time; i++)
  {
    __delay_cycles(INVOKE_DELAY_1MS);
  }
}

/*** BSL entry sequence for Legacy BSL ******************************************************/
void BSL_legacy_invoke_sequence()
{ 
     /* BSL invoke sequence -- fix
     *
     *       H -----------+     +--+  +-----+
     * TEST  L            +-----+  +--+     +--------
     *
     *       H --------------+           +-----------
     * RST   L               +-----------+
     *
     *
     *       H -----------------+  +--+     +--------
     * TCK   L                  +--+  +-----+
     */

    // Set RST, TST and TCK pin to output
    ENTRY_SEQ_PDIR |= (RESET_PIN + TEST_PIN + TCK_PIN);

    /* (1)RST low, TEST low, TCK low */
    ENTRY_SEQ_POUT = 0;
    BSL_delay_ms(DELAY_12MS);
    BSL_delay_ms(DELAY_12MS);

    /* (2)RST low, TEST low, TCK high */
    ENTRY_SEQ_POUT = TCK_PIN;
    BSL_delay_ms(DELAY_12MS);

    /* (3)RST low, TEST high, TCK low*/
    ENTRY_SEQ_POUT = TEST_PIN ;
    BSL_delay_ms(DELAY_12MS);

    /* (4)RST low, TEST low, TCK high*/
    ENTRY_SEQ_POUT = TCK_PIN;
    BSL_delay_ms(DELAY_12MS);

    /* (5)RST low, TEST high, TCK low */
    ENTRY_SEQ_POUT =TEST_PIN;
    BSL_delay_ms(DELAY_12MS);

    /* (6)RST high, TEST high, TCK low */
    ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN;
    BSL_delay_ms(DELAY_12MS);
    
    /* (7)RST high, TEST low, TCK high */
    ENTRY_SEQ_POUT = RESET_PIN + TCK_PIN;
    BSL_delay_ms(DELAY_12MS);

    // Set RST, TST and TCK pin to input
    ENTRY_SEQ_PDIR &= ~(RESET_PIN + TEST_PIN + TCK_PIN);
}

/*** BSL exit sequence (JTAG20 workaround) ************************************/

void BSL_reset_sequence()
{
  /* BSL reset sequence (JTAG20 workaround)
   *
   *       H -------------------------+  +----
   * RST   L                          +--+
   *                            >|50us|<
   *       H --------+  +--+  +--+
   * TEST  L         +--+  +--+  +------------
   */

  // Set RST, TST and TCK pin to output
  ENTRY_SEQ_PDIR |= (RESET_PIN + TEST_PIN + TCK_PIN);

  // (1)Start with RST and TEST high 
  ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (2)RST high, TEST low
  ENTRY_SEQ_POUT = RESET_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (3)RST high, TEST high
  ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (4)RST high, TEST low
  ENTRY_SEQ_POUT = RESET_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (5)RST high, TEST high
  ENTRY_SEQ_POUT = RESET_PIN + TEST_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (6)RST high, TEST low
  ENTRY_SEQ_POUT = RESET_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // (7)RST low, TEST low
  ENTRY_SEQ_POUT = 0;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);
  
  // (8)RST high, TEST low
  ENTRY_SEQ_POUT = RESET_PIN;
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);
  __delay_cycles(BSL_EXIT_SEQUENCE_TIME);

  // Set RST, TST and TCK pin to input
  ENTRY_SEQ_PDIR &= ~(RESET_PIN + TEST_PIN + TCK_PIN);
}

/*** Baudrate Select handles configuration of UART, I2C and SPI peripheral 
     on baud rate change detected by USBCDC_handleSetLineCoding() *************/

uint8_t isDefaultI2cAddr = 1;   // Variable that assign status is default device I2C addressing is chosen
uint8_t is7bitI2cAddr = 0;      // Variable that assign status is custom 7-bit I2C addressing is chosen
uint8_t is10bitI2cAddr = 0;     // Variable that assign status is custom 10-bit I2C addressing is chosen
uint16_t customI2cAddr = 0;     // Variable that assign slave address contains in the baudrate
uint8_t isI2c100kHzBitrate = 0; // Variable that assign if the custom 100kHz bitrate is chosen

uint8_t BaudrateSelect(uint32_t lBaudrate)
{
  if(current_baudrate == lBaudrate)
  {
      return 0;
  }

  current_baudrate = lBaudrate;

  // Reset peripheral communication pins to input
  P4DIR &= ~(BIT0 + BIT3 + BIT4 + BIT5);
  // Set RST, TST and TCK pin to input
  ENTRY_SEQ_PDIR &= ~(RESET_PIN + TEST_PIN + TCK_PIN);

  LED1_OFF
  __delay_cycles(1000000);
  LED1_ON

  uint8_t baudIndex = 0;

  switch(lBaudrate)
  {
    // UART Peripheral
    case BAUDRATE_1200:
    case BAUDRATE_2400:
    case BAUDRATE_4800:
    case BAUDRATE_9600:
    case BAUDRATE_19200:
    case BAUDRATE_38400:
    case BAUDRATE_57600:
    case BAUDRATE_115200:
    case BAUDRATE_230400:
    case BAUDRATE_460800:
    case BAUDRATE_921600:
    case BAUDRATE_INVOKE_ROCKET:
    case BAUDRATE_RESET_SEQUENCE:
    case BAUDRATE_INVOKE_MSP430:
    case BAUDRATE_INVOKE_LEGACY_MSP430:
    case BAUDRATE_EXIT_MSP432E4XX:
    case BAUDRATE_INVOKE_9600_MSP432E4XX:
    case BAUDRATE_INVOKE_19200_MSP432E4XX:
    case BAUDRATE_INVOKE_38400_MSP432E4XX:
    case BAUDRATE_INVOKE_57600_MSP432E4XX:
    case BAUDRATE_INVOKE_115200_MSP432E4XX:
      active_peripheral = UART;
      baudIndex = InitUart(lBaudrate);
      break;

    // I2C Peripheral
    case BAUDRATE_100KHZ:
    case BAUDRATE_400KHZ:
    case BAUDRATE_INVOKE_100KHZ_MSP430:
    case BAUDRATE_INVOKE_400KHZ_MSP430:
    case BAUDRATE_INVOKE_100KHZ_MSP432E4XX:
    case BAUDRATE_INVOKE_400KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_100KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_400KHZ_MSP432E4XX:
      active_peripheral = I2C;
      baudIndex = InitI2C(lBaudrate);
      break;

    // SPI peripheral
    case BAUDRATE_125KHZ:
    case BAUDRATE_250KHZ:
    case BAUDRATE_500KHZ:     
    case BAUDRATE_1MHZ:
    case BAUDRATE_INVOKE_125KHZ_MSP430:
    case BAUDRATE_INVOKE_250KHZ_MSP430:
    case BAUDRATE_INVOKE_500KHZ_MSP430:
    case BAUDRATE_INVOKE_1MHZ_MSP430:
    case BAUDRATE_INVOKE_125KHZ_MSP432E4XX:
    case BAUDRATE_INVOKE_250KHZ_MSP432E4XX:
    case BAUDRATE_INVOKE_500KHZ_MSP432E4XX:
    case BAUDRATE_INVOKE_1MHZ_MSP432E4XX:
    case BAUDRATE_EXIT_125KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_250KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_500KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_1MHZ_MSP432E4XX:      
      active_peripheral = SPI;
      baudIndex = spiInit(lBaudrate);
      break;

    default:
      isI2c100kHzBitrate = 0;
      isDefaultI2cAddr = 0;  
      //get the 100kHz I2C range and also 7-bit slave addressing value
      if((lBaudrate > BAUDRATE_100KHZ) && (lBaudrate <= (BAUDRATE_100KHZ+MAX_SLAVE_ADDR_7BIT)))
      {
        active_peripheral = I2C;
        is7bitI2cAddr = 1;  
        is10bitI2cAddr = 0;
        customI2cAddr = (uint16_t)(MAX_SLAVE_ADDR_7BIT&(lBaudrate-BAUDRATE_100KHZ));
        isI2c100kHzBitrate = 1;
        baudIndex = InitI2C(lBaudrate);
      }
      //get the 100kHz I2C range and also 10-bit slave addressing value
      else if((lBaudrate > BAUDRATE_100KHZ) && (lBaudrate <= (BAUDRATE_100KHZ+MAX_SLAVE_ADDR_10BIT)))
      {
        active_peripheral = I2C;
        is7bitI2cAddr = 0;  
        is10bitI2cAddr = 1;
        customI2cAddr = (uint16_t)(MAX_SLAVE_ADDR_10BIT&(lBaudrate-BAUDRATE_100KHZ));
        isI2c100kHzBitrate = 1;
        baudIndex = InitI2C(lBaudrate);
      }
      //get the 400kHz I2C range and also 7-bit slave addressing value
      else if((lBaudrate > BAUDRATE_400KHZ) && (lBaudrate <= (BAUDRATE_400KHZ+MAX_SLAVE_ADDR_7BIT)))
      {
        active_peripheral = I2C;
        is7bitI2cAddr = 1;  
        is10bitI2cAddr = 0;
        customI2cAddr = (uint16_t)(MAX_SLAVE_ADDR_7BIT&(lBaudrate-BAUDRATE_400KHZ));
        baudIndex = InitI2C(lBaudrate);
      }
      //get the 400kHz I2C range and also 10-bit slave addressing value
      else if((lBaudrate > BAUDRATE_400KHZ) && (lBaudrate <= (BAUDRATE_400KHZ+MAX_SLAVE_ADDR_10BIT)))
      {
        active_peripheral = I2C;
        is7bitI2cAddr = 0;  
        is10bitI2cAddr = 1;
        customI2cAddr = (uint16_t)(MAX_SLAVE_ADDR_10BIT&(lBaudrate-BAUDRATE_400KHZ));
        baudIndex = InitI2C(lBaudrate);
      }
      break;
  }
  
  return baudIndex;
}


/*** UART ********************************************************************/

// List of standard baudrates
const struct _BaudrateList {uint8_t ucaBR0, ucaBR1, ucaMCTL;} // CLK 20MHz
  BaudrateList[] = 
  {                                   
    {0x00,    0x00,   0x00},
    {0x1A,    0x41,   UCBRS_5},                   // 1.2 kb/s
    {0x8D,    0x20,   UCBRS_3},                   // 2.4 kb/s
    {0x46,    0x10,   UCBRS_5},                   // 4.8 kb/s
    {0x82,    0x00,   0x30 + UCOS16},             // 9.6 kb/s
    {0x41,    0x00,   (UCBRF_2+UCOS16)},          // 19.2 kb/s
    {0x20,    0x00,   (UCBRF_9+UCOS16)},          // 38.4 kb/s
    {0x15,    0x00,   (UCBRF_11+UCOS16)},         // 57.6 kb/s
    {0x0A,    0x00,   (UCBRF_14+UCOS16)},         // 115.2 kbit/s
    {0x05,    0x00,   (UCBRF_7+UCOS16)},          // 230.4 kbit/s
    {0x02,    0x00,   (UCBRS_6+UCBRF_10+UCOS16)}, // 460.8 kbit/s
    {0x15,    0x00,   (UCBRS_6)                   // 921.6 kbit/s
  }                  
};

/*** Initialization UART module that will be called everything the baud rate is
     changed*******************************************************************/

uint8_t InitUart(uint32_t lBaudrate)
{
  uint8_t baudIndex = BAUD_ERROR_PARAMS;

  //Simple port mapping
  __disable_interrupt();                  // Disable Interrupts before altering Port Mapping registers
  PMAPKEYID = PMAPKEY;
  P4MAP0 = PM_NONE;
  P4MAP3 = PM_NONE;
  P4MAP4 = PM_UCA0TXD;
  P4MAP5 = PM_UCA0RXD;
  PMAPKEYID = 0;
  __enable_interrupt();

  P4SEL |= BIT0 + BIT3 + BIT4 + BIT5;

  // configure USCI_A0 UART
  UCA0CTL1 |= UCSWRST;                    // Put state machine in reset
  UCA0CTL1 |= UCSSEL__SMCLK;              // SMCLK
  
  switch(lBaudrate)
  {
    #if (RX_BUFFER_SIZE > (BAUDRATE_1200*SYS_DELAY/1000/8))
    case BAUDRATE_1200: 
      baudIndex = BAUD_1200_PARAMS; 
      break;
    #endif
          
    #if (RX_BUFFER_SIZE > (BAUDRATE_2400*SYS_DELAY/1000/8))
    case BAUDRATE_2400: 
      baudIndex = BAUD_2400_PARAMS; 
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_4800*SYS_DELAY/1000/8))
    case BAUDRATE_4800: 
      baudIndex = BAUD_4800_PARAMS; 
      break;
    #endif
          
    case BAUDRATE_INVOKE_ROCKET:         // Invoke Rocket's BSL
        __disable_interrupt();
        USB_disable();
        __delay_cycles(500000);
        ((void(*)(void))BSL_ADDRESS)();  // Call #0x1000, enter BSL
        break;

    case BAUDRATE_RESET_SEQUENCE:
        BSL_reset_sequence();
        baudIndex = BAUD_9600_PARAMS;
        break;
        
    #if(RX_BUFFER_SIZE > (BAUDRATE_9600*SYS_DELAY/1000/8))
    case BAUDRATE_9600: 
      baudIndex = BAUD_9600_PARAMS; 
      break;
    #endif
      
    case BAUDRATE_INVOKE_MSP430:
      BSL_invoke_sequence();
      baudIndex = BAUD_9600_PARAMS;
      break;
      
    case BAUDRATE_INVOKE_LEGACY_MSP430:
      BSL_legacy_invoke_sequence();
      baudIndex = BAUD_9600_PARAMS;
      break;

    case BAUDRATE_INVOKE_9600_MSP432E4XX:
      isE4xx = 1;
      baudIndex = BAUD_9600_PARAMS;
      break;
      
    case BAUDRATE_INVOKE_19200_MSP432E4XX:
      isE4xx = 1;
      baudIndex = BAUD_19200_PARAMS;
      break;
      
    case BAUDRATE_INVOKE_38400_MSP432E4XX:
      isE4xx = 1;
      baudIndex = BAUD_38400_PARAMS;
      break;
      
    case BAUDRATE_INVOKE_57600_MSP432E4XX:
      isE4xx = 1;
      baudIndex = BAUD_57600_PARAMS;
      break;
      
    case BAUDRATE_INVOKE_115200_MSP432E4XX:
      isE4xx = 1;
      baudIndex = BAUD_115200_PARAMS;
      break;
      
    case BAUDRATE_EXIT_MSP432E4XX:
      isE4xx = 0;
      baudIndex = BAUD_9600_PARAMS;
      break;
            
    #if(RX_BUFFER_SIZE > (BAUDRATE_19200*SYS_DELAY/1000/8))
    case BAUDRATE_19200: 
      baudIndex = BAUD_19200_PARAMS; 
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_38400*SYS_DELAY/1000/8))
    case BAUDRATE_38400: 
      baudIndex = BAUD_38400_PARAMS; 
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_57600*SYS_DELAY/1000/8))
    case BAUDRATE_57600: 
      baudIndex = BAUD_57600_PARAMS;
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_115200*SYS_DELAY/1000/8))
    case BAUDRATE_115200: 
      baudIndex = BAUD_115200_PARAMS; 
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_230400*SYS_DELAY/1000/8))
    case BAUDRATE_230400: 
      baudIndex = BAUD_230400_PARAMS; 
      break;
    #endif
          
    #if(RX_BUFFER_SIZE > (BAUDRATE_460800*SYS_DELAY/1000/8))
    case BAUDRATE_460800:
      baudIndex = BAUD_460800_PARAMS; 
      break;
    #endif
        
    #if(RX_BUFFER_SIZE > (BAUDRATE_921600*SYS_DELAY/1000/8))
    case BAUDRATE_921600: 
      baudIndex = BAUD_921600_PARAMS; 
      break;
    #endif
        
    default: 
      baudIndex = BAUD_ERROR_PARAMS; 
      break;
  }
  
  // MSP432E UART bootloader configured with parity disabled
  if (isE4xx)
  {
    UCA0CTL0 = 0;
  }
  // MSP432P and MSP430 UART bootloader configured with even-parity enabled
  else
  {
    UCA0CTL0 = UCPEN + UCPAR;
  }
  
  if(baudIndex != BAUD_ERROR_PARAMS)
  {
    UCA0BR0 = BaudrateList[baudIndex].ucaBR0;
    UCA0BR1 = BaudrateList[baudIndex].ucaBR1;
    UCA0MCTL = BaudrateList[baudIndex].ucaMCTL;
  }

  UCA0CTL1 &= ~UCSWRST; // Initialize USCI state machine
  UCA0IE |= UCRXIE;
  return baudIndex;
}

/*** main function for UART bootloader that will be looped in main function 
     once UART peripheral is initialized***************************************/

void uartControl()
{
  // Send data from USB buffer to UART
  if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
  {
    if(UCA0IFG & UCTXIFG)
    {
      USBCDC_receiveData((uint8_t *)&UCA0TXBUF, 1, CDC0_INTFNUM);
      LED2_TOGGLE
    }
  }

  if(UARTtoCDC_TXPtr != UARTtoCDC_RXPtr)
  {
    while(cdcSendDataWaitTilDone((uint8_t*)&UARTtoCDC_CircularBuffer[UARTtoCDC_TXPtr], 1, CDC0_INTFNUM, UART_TIMEOUT_CYCLES) != 0);
    UARTtoCDC_TXPtr++;
    if (UARTtoCDC_TXPtr == BSL_BUFFER_SIZE)
    {
        UARTtoCDC_TXPtr = 0;
    }
  }
}

/***Interrupt Service Routine for UART peripheral *****************************/

#pragma vector=USCI_A0_VECTOR
__interrupt void USCI_A0_ISR(void)
{
  switch(__even_in_range(UCA0IV,4))
  {
    case 0:                               // Vector 0 - no interrupt
      break;
      
    case 2:                               // Vector 2 - RXIFG
      // Store received byte in TX buffer
      UARTtoCDC_CircularBuffer[UARTtoCDC_RXPtr++] = UCA0RXBUF;
      if (UARTtoCDC_RXPtr == BSL_BUFFER_SIZE)
      {
          UARTtoCDC_RXPtr = 0;
      }
      break;
     
    case 4:	                              // Vector 4 - TXIFG
      break;
   
    default: 
      break;
  }
}

/*** I2C *********************************************************************/

unsigned char ACK = 0;              // Variable that shows is data received 
                                    // from BSL target device
unsigned char state = SWAIT;        // Variable of I2C FSM state

unsigned int BSLCommandLength = 0;  // Total bytes received from USBCDC to be
                                    // transmitted via I2C to the BSL target device
unsigned int rxlength = 0;          // Number received bytes from BSL target
                                    // device calculated from BSL response
                                    // for MSP430 and MSP432P 
unsigned int txCmdLength = 0;       // Expected received BSL command bytes from USBCDC
unsigned int isGetStatusE4xx = 0;   // Variable to acknowledge Get Status command
                                    // for MSP432E

int16_t dataToUSBCDCLength = 0;     // Total bytes received from BSL target device
                                    // that will be transmitted back to USBCDC
int16_t requireSlaveAnswer = 1;     // Expected received bytes from BSL target
                                    // device as BSL response

extern volatile unsigned char bDataReceiveCompleted_event[]; // data receive completed event
volatile unsigned char bDataSendCompleted_event[CDC_NUM_INTERFACES] = {FALSE};
volatile unsigned char bDataReceiveCompleted_event[CDC_NUM_INTERFACES] = {FALSE}; 

unsigned char i2cRxBuffer[BSL_BUFFER_SIZE]; // Buffer to store received bytes from BSL target device
unsigned char i2cTxBuffer[BSL_BUFFER_SIZE]; // Buffer to store BSL command from USBCDC
                                            // that will be transmitted to I2C to the BSL target device

unsigned char *PTxData; // Pointer to i2cTxBuffer
unsigned char *PRxData; // Pointer to i2cRxBuffer
int TXByteCtr;          // Number of bytes that been transmitted from i2cTxBuffer to the BSL target device
int RXByteCtr;          // Number of bytes that been received from BSL target device

enum 
{
  I2C_IDLE,
  I2C_RECEIVING,
  I2C_RECEIVE_COMPLETE,
  I2C_TRANSMITTING,
  I2C_TRANSMIT_COMPLETE,
  I2C_ERROR
} I2C_Status_e;


/*** Initialization of the USCI Module for I2C ********************************/

int8_t InitI2C(uint32_t bitrate)
{
  int8_t ret = 1;

  //Simple port mapping
  __disable_interrupt(); // Disable Interrupts before altering Port Mapping registers
  PMAPKEYID = PMAPKEY;
  P4MAP0 = PM_UCB0SCL;
  P4MAP3 = PM_NONE;
  P4MAP4 = PM_NONE;
  P4MAP5 = PM_UCB0SDA;
  PMAPKEYID = 0;
  __enable_interrupt();

  P4SEL |= BIT0 + BIT3 + BIT4 + BIT5;

  UCB0CTL1 = UCSWRST;                         // Enable SW reset

  UCB0CTL0 = UCMST + UCMODE_3 + UCSYNC;       // I2C Master, synchronous mode
  UCB0CTL1 = UCSSEL_2 + UCSWRST;              // Use SMCLK, keep SW reset

  switch(bitrate)
  {
    case BAUDRATE_INVOKE_100KHZ_MSP430:
        BSL_invoke_sequence();
    case BAUDRATE_INVOKE_100KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_100KHZ_MSP432E4XX:
        LED1_TOGGLE
        LED2_TOGGLE
        LED2_ON
    case BAUDRATE_100KHZ:      
        UCB0BR0 = SCL_CLOCK_DIV(BAUDRATE_100KHZ);   // set prescaler
        break;
   
    case BAUDRATE_INVOKE_400KHZ_MSP430:
        BSL_invoke_sequence();
    case BAUDRATE_INVOKE_400KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_400KHZ_MSP432E4XX:
        LED1_TOGGLE
        LED2_TOGGLE
        LED2_ON
    case BAUDRATE_400KHZ:      
        UCB0BR0 = SCL_CLOCK_DIV(BAUDRATE_400KHZ);   // set prescaler
        break;
   
    default:
        if(isI2c100kHzBitrate)
        {
           UCB0BR0 = SCL_CLOCK_DIV(BAUDRATE_100KHZ); 
        }
        else
        {
           UCB0BR0 = SCL_CLOCK_DIV(BAUDRATE_400KHZ); 
        }
        break;
  }

  // enable isE4xx variable when it is invoked
  if ((bitrate==BAUDRATE_INVOKE_100KHZ_MSP432E4XX) || 
      (bitrate==BAUDRATE_INVOKE_400KHZ_MSP432E4XX))
  {
    isE4xx=1;
  }
  
  // disable isE4xx variable when it is exited
  if ((bitrate==BAUDRATE_EXIT_100KHZ_MSP432E4XX) || 
      (bitrate==BAUDRATE_EXIT_400KHZ_MSP432E4XX) ||
      (bitrate==BAUDRATE_INVOKE_100KHZ_MSP430)   ||
      (bitrate==BAUDRATE_INVOKE_400KHZ_MSP430))
  {
    isE4xx=0;
  }

  UCB0BR1 = 0;

  // Set I2C slave address
  if(isDefaultI2cAddr) 
  {
    if (isE4xx)
    {
      UCB0I2CSA = BSL_SLAVE_ADDR_MSP432E4xx;    // Set slave address for MSP432E4xx        
    }
    else
    {
      UCB0I2CSA = BSL_SLAVE_ADDR;               // Set slave address for MSP430 and MSP432P 
    }
  }
  else
  {
    // Assign custom I2C address
    if (is10bitI2cAddr)
    {
      UCB0CTL0 |= UCSLA10;
    }
    UCB0I2CSA = customI2cAddr;
  }
 
  I2C_PORT_SEL |= SDA_PIN + SCL_PIN;          // select module function for the used I2C pins

  UCB0CTL1 &= ~UCSWRST;                       // Clear SW reset, resume operation

  PRxData = i2cRxBuffer;                      // In case no receive buffer is assigned

  if(UCB0STAT & UCBBUSY)                      // test if bus to be free
  {                                           // otherwise a manual Clock on is generated
      I2C_PORT_SEL &= ~SCL_PIN;               // Select Port function for SCL
      I2C_PORT_OUT &= ~SCL_PIN;               //
      I2C_PORT_DIR |= SCL_PIN;                // drive SCL low
      I2C_PORT_SEL |= SDA_PIN + SCL_PIN;      // select module function for the used I2C pins
  };

  I2C_Status_e = I2C_IDLE;
  return ret;
}

/*** Send the message that is received from USBCDC to I2CTXBUF ****************/

int16_t i2cSendMessage(unsigned char* I2cMessage, int messageLength)
{
  UCB0IE &= ~UCRXIE;                            // disable RX ready interrupt
  UCB0IFG &= ~(UCTXIFG + UCSTPIFG + UCNACKIFG); // clear TX ready and stop interrupt flag
  UCB0IE |= UCTXIE | UCNACKIE;                  // enable TX ready interrupt

  PTxData = (unsigned char *)I2cMessage;        // TX array start address

  TXByteCtr = messageLength ;                   // Load TX byte counter

  I2C_Status_e = I2C_TRANSMITTING;
  UCB0CTL1 |= UCTR + UCTXSTT;                   // I2C TX, start condition

  while(I2C_Status_e == I2C_TRANSMITTING);

  if(I2C_Status_e == I2C_TRANSMIT_COMPLETE)
  {
    PTxData = (unsigned char *)I2cMessage;        // TX array start address

    if (isE4xx)
    {
      // for MSP432E4xx, the number of response bytes are fix
      // if it is a GET_STATUS command, the number of response byte is 5 bytes
      // other than that, the number of response byte is 2 bytes
      // this variable will be used in i2cReceiveMessage
      if (*(PTxData + BSL_LENGTH_LENGTH) == BSL_CMD_GET_STATUS)
      {
        isGetStatusE4xx = 1;
      }
      else
      {
        isGetStatusE4xx = 0;
      }
    }
    
    // check for commands that does not require any ACK from BSL target
    // RX_DATA_BLOCK_FAST for MSP430 and MSP432P
    // ACK sign from host to the BSL for MSP432E
    if((*(PTxData + BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH) == BSL_CMD_RX_BLOCK_FAST) ||
        ((*(PTxData) == BSL_RESPONSE_ACK) && (*(PTxData+ BSL_HEADER_LENGHT) == BSL_RESPONSE_ACK_E4xx)))
    {
      return 0;
    }
    // for MSP430 and MSP432P: the exact number of response will be calculated
    // from the received response bytes. 
    // for MSP432E the exact number of response will be calculated based on
    // isGetStatusE4xx value.
    // The return '1' below is to invoke the i2cReceivedMessage to receive the  
    // first ACK byte and proceed with the FSM
    else
    {
      return 1;
    }  
  }
  else if(I2C_Status_e == I2C_ERROR)
  {
      return RET_ERROR;
  }
  return RET_ERROR;
}

/*** Received the response bytes from BSL target device, keep it in a 
     buffer until all bytes received completedly, before transmitting it back 
     to the USBCDC*************************************************************/

int16_t i2cReceiveMessage(unsigned char* I2cMessage)
{
  uint32_t timeout = TIMEOUT_RECEPTION;

  UCB0CTL1 &= ~UCTR;
  UCB0IE |= UCRXIE;                         // Enable RX interrupt

  PRxData = (unsigned char *)I2cMessage;    // Start of RX buffer
  RXByteCtr = 1;                            // Load RX byte counter

  ACK = 0;
  I2C_Status_e = I2C_RECEIVING;
  UCB0CTL1 |= UCTXSTT;

  while((I2C_Status_e == I2C_RECEIVING) && (timeout-- != 0));

  if((timeout == 0) || (I2C_Status_e == I2C_ERROR))
  {
      UCB0CTL1 |= UCTXSTP;
      timeout = TIMEOUT_RECEPTION;
      while((UCB0CTL1 & UCTXSTP) && (timeout-- != 0));
  }

  if(I2C_Status_e == I2C_RECEIVE_COMPLETE)
  {
    if (isE4xx)
    {
      if(isGetStatusE4xx)
      {
        return BSL_E4xx_RESPONSE_GET_STATUS;
        
      }
      else
      {
        return BSL_E4xx_RESPONSE_ACK;
      }
    }
    else
    {
      if (rxlength > 0)
      {
        return (BSL_ACK_LENGTH + BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH + rxlength + BSL_CRC_LENGTH);
      }
      else
      {
        return (BSL_ACK_LENGTH + BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH);
      }
    }
  }
  else
  {
    return RET_ERROR;
  }
}

/*** main function for I2C bootloader that will be looped in main function 
     once I2C peripheral is initialized***************************************/

void i2cControl()
{
  unsigned int bytes_sent, bytes_received;
  unsigned char ret = USBCDC_intfStatus(CDC0_INTFNUM, &bytes_sent,  &bytes_received);
  
  if (ret & kUSBCDC_dataWaiting)
  {
    BSLCommandLength = i2cFsm(i2cTxBuffer);
  }
  if (BSLCommandLength > 0)
  {
    // Special command to just get data
    if (BSLCommandLength == BSL_CUSTOM_CMD_LENGTH)
    {
        // If response is OK, request response from Slave
        dataToUSBCDCLength = i2cReceiveMessage(i2cTxBuffer);
        if ((requireSlaveAnswer < 0) || (dataToUSBCDCLength < 0))
        {
            // If there was an error, report to PC
            dataToUSBCDCLength = BSL_CUSTOM_CMD_LENGTH;
            i2cTxBuffer[0] = BSL_UNKNOWN_ERROR_CODE;   // Temporary error code
        }
    }
    else
    {
      requireSlaveAnswer = i2cSendMessage(i2cTxBuffer, BSLCommandLength);
      UCB0CTL1 |= UCTXSTP;
      __delay_cycles(I2C_DELAY_CYCLES);

      if (requireSlaveAnswer == 0)
      {
          dataToUSBCDCLength = 0; // Don't send response to PC
      }
      else if(requireSlaveAnswer > 0)
      {
          // If response is OK, request response from Slave
          dataToUSBCDCLength = i2cReceiveMessage(i2cRxBuffer);
      }

      if ((requireSlaveAnswer < 0) || (dataToUSBCDCLength < 0))
      {
          // If there was an error, report to PC
          dataToUSBCDCLength = BSL_CUSTOM_CMD_LENGTH;
          i2cTxBuffer[0] = BSL_UNKNOWN_ERROR_CODE;   // Error code
      }
      BSLCommandLength = 0;
    }
  }

  if (dataToUSBCDCLength > 0)
  {
    //send data back to PC
    bDataSendCompleted_event[0] = FALSE;
    ret = USBCDC_sendData((unsigned char*)&i2cRxBuffer, dataToUSBCDCLength, CDC0_INTFNUM);
    while (bDataSendCompleted_event == FALSE){};
    dataToUSBCDCLength = 0;
  }
  isGetStatusE4xx = 0;
}

/***  I2C FSM to wait BSL command received from USBCDC*************************/

unsigned int i2cFsm(unsigned char* dataBuffer)
{
  unsigned int retvalue = 0;
  
  switch(state)
  {
    case(SWAIT):	
      // wait for UART_HEADER to start the data sequence
      USBCDC_receiveData(dataBuffer, BSL_HEADER_LENGHT, CDC0_INTFNUM);
      
      // waiting for the header MSP430 and MSP432P packet
      if(dataBuffer[0] == BSL_STD_CMD_HEADER)
      {
        txCmdLength = 0;
        state = SSTART;
      }
      // waiting for the header custom command packet
      else if (dataBuffer[0] == BSL_CUSTOM_CMD_HEADER)
      {
        state = SCMD;
      }
      // waiting for MSP432E4xx host ACK response
      else if ((isE4xx==1) && (dataBuffer[0] == BSL_RESPONSE_ACK))
      {
        state  = SDATA_E4XX;
        txCmdLength = BSL_ACK_LENGTH;
      }
      // waiting for other MSP432E4xx first command byte which is the number
      // of bytes for BSL command
      else
      {
        state  = SDATA_E4XX;
        txCmdLength = dataBuffer[0] - BSL_ACK_LENGTH;
      }
      retvalue = 0;
      break;
    
    // waiting for all data being sent from the host for MSP432E4xx command 
    case(SDATA_E4XX):
      bDataReceiveCompleted_event[0] = FALSE;
      USBCDC_receiveData(dataBuffer + BSL_HEADER_LENGHT, txCmdLength, CDC0_INTFNUM);
      while (bDataReceiveCompleted_event[0] == FALSE){};
      USBCDC_rejectData(CDC0_INTFNUM);	// discard leftover bytes
      state = SWAIT;
      retvalue = txCmdLength+1;
      break;
        
    // waiting for number of bytes information from the host for MSP430 and MSP432P command 
    case(SSTART):	// read in length bytes
      bDataReceiveCompleted_event[0] = FALSE;
      USBCDC_receiveData(dataBuffer + BSL_HEADER_LENGHT, BSL_LENGTH_LENGTH, CDC0_INTFNUM);
      while (bDataReceiveCompleted_event[0] == FALSE){}; // wait until data received
      txCmdLength = ((unsigned char) dataBuffer[1 + BSL_HEADER_LENGHT] << 8) | dataBuffer[0 + BSL_HEADER_LENGHT];
      state = SDATA;
      retvalue = 0;
      break;
    
    // waiting for the rest of data (+ crc) from the host for MSP430 and MSP432P command 
    case(SDATA):
      bDataReceiveCompleted_event[0] = FALSE;
      USBCDC_receiveData(dataBuffer + BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH, txCmdLength + BSL_CRC_LENGTH, CDC0_INTFNUM);
      while (bDataReceiveCompleted_event[0] == FALSE){};
      USBCDC_rejectData(CDC0_INTFNUM);		// discard leftover bytes
      state = SWAIT;
      retvalue = BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH + txCmdLength + BSL_CRC_LENGTH;
      break;

    // waiting for the rest of data for custom command 
    case SCMD:
      USBCDC_receiveData(dataBuffer, BSL_HEADER_LENGHT, CDC0_INTFNUM);
      retvalue = dataBuffer[0];
      state = SWAIT;
      break;

    default:
      state = SWAIT;
      retvalue = 0;
      break;
  }
  return retvalue;
}

/***Interrupt Service Routine for I2C peripheral *****************************/

#pragma vector = USCI_B0_VECTOR
__interrupt void USCI_B0_ISR(void)
{
  switch(__even_in_range(UCB0IV,USCI_I2C_UCTXIFG))
  {
    case USCI_NONE:                 // Vector 0:  No interrupts
      break; 
      
    case USCI_I2C_UCALIFG:          // Vector 2:  UCALIFG  
      break; 
      
    case USCI_I2C_UCNACKIFG:
        I2C_Status_e = I2C_ERROR;   // Vector 4:  NACKIFG
        break;                  
        
    case USCI_I2C_UCSTTIFG:         // Vector 6:  STTIFG
      break; 
      
    case USCI_I2C_UCSTPIFG:         // Vector 8:  STPIFG
      break; 
      
    case USCI_I2C_UCRXIFG:          // Vector 10: RXIFG
      if (ACK == 0)
      {
          *PRxData++ = UCB0RXBUF; // Move RX data to address PRxData
          if(*(PRxData-1) != BSL_RESPONSE_ACK)
          {
            UCB0CTL1 |= UCTXSTP;  // Generate I2C stop condition
          }
          else
          {
            ACK = 1;
            if (isE4xx)
            {
              if (isGetStatusE4xx)
              {
                RXByteCtr = (BSL_E4xx_RESPONSE_GET_STATUS-BSL_ACK_LENGTH); // receives the next four bytes
              }
              else
              {
                // last byte received requires stop condition from the master
                UCB0CTL1 |= UCTXSTP; 
                RXByteCtr = BSL_ACK_LENGTH; // receives the next one byte
              }               
            }
            else
            {
              RXByteCtr = BSL_HEADER_LENGHT + BSL_LENGTH_LENGTH;  // receives the next three bytes
              rxlength = 0;          
            }

          }
      }
      else
      {
        RXByteCtr--;                                // Decrement RX byte counter
        if(RXByteCtr)
        {
          *PRxData++ = UCB0RXBUF;                   // Move RX data to address PRxData
          if (isE4xx)
          {
            if((RXByteCtr == 1))                    // Only one byte left?
            {
                UCB0CTL1 |= UCTXSTP;                // Generate I2C stop condition        
            }
          }
          else
          {
            if((RXByteCtr == 1) && (rxlength > 0))  // Only one byte left?
            {
                UCB0CTL1 |= UCTXSTP;                // Generate I2C stop condition
            }            
          }
        }
        else if((isE4xx==0)&&(rxlength == 0)) 
        {
          *PRxData++ = UCB0RXBUF;                   // Move RX data to address PRxData
          rxlength = ((unsigned int) *(PRxData-1) << 8) | *(PRxData-2);
          if(rxlength == 0) UCB0CTL1 |= UCTXSTP;    // stop receiving if no bytes required
          {
              RXByteCtr = rxlength + BSL_CRC_LENGTH;
          }
        }
        else
        {
          *PRxData = UCB0RXBUF;                     // Move final RX data to PRxData
          I2C_Status_e = I2C_RECEIVE_COMPLETE;
        }
      }
      break;
      
    case USCI_I2C_UCTXIFG:                          // Vector 12: TXIFG
      if(TXByteCtr)                                 // Check TX byte counter
      {
          UCB0TXBUF = *PTxData++;                   // Load TX buffer
          TXByteCtr--;                              // Decrement TX byte counter
      }
      else
      {
          I2C_Status_e = I2C_TRANSMIT_COMPLETE;
      }
      break;
        
    default:
        break;
  }
}


/*** SPI *********************************************************************/
unsigned int isHostAck = 0;

/***Initialize SPI peripheral, this function called when the baudrate change is 
    detected and the Baudrate Select function assign SPI as ctive peripheral***/
uint8_t spiInit(uint32_t lBaudrate)
{
  //Simple port mapping
  __disable_interrupt();                  // Disable Interrupts before altering Port Mapping registers
  PMAPKEYID = PMAPKEY;
  P4MAP0 = PM_UCB1CLK;
  P4MAP3 = PM_NONE;  // STE is driven by I/O pin (see below)
  P4MAP4 = PM_UCB1SIMO;
  P4MAP5 = PM_UCB1SOMI;
  PMAPKEYID = 0;
  __enable_interrupt();

  P4SEL = BIT0 + BIT4 + BIT5;

  // Use P4.3 as (active low) STE signal
  P4DIR |= BIT3;
  P4OUT |= BIT3;  // Disable SPI slave

  // configure USCI_B1 SPI
  UCB1CTL1 = UCSWRST;                     // **Put state machine in reset**
  UCB1CTL1 |= UCSSEL__ACLK;

  switch(lBaudrate)
  {
    case BAUDRATE_INVOKE_125KHZ_MSP430:
      BSL_invoke_sequence();
    case BAUDRATE_INVOKE_125KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_125KHZ_MSP432E4XX:
    case BAUDRATE_125KHZ:
      UCB1BRW = 32;
      break;
      
    case BAUDRATE_INVOKE_250KHZ_MSP430:
      BSL_invoke_sequence();
    case BAUDRATE_INVOKE_250KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_250KHZ_MSP432E4XX:
    case BAUDRATE_250KHZ:
      UCB1BRW = 16;
      break;
      
    case BAUDRATE_INVOKE_500KHZ_MSP430:
      BSL_invoke_sequence();
    case BAUDRATE_INVOKE_500KHZ_MSP432E4XX:
    case BAUDRATE_EXIT_500KHZ_MSP432E4XX:
    case BAUDRATE_500KHZ:
      UCB1BRW = 8;
      break;
      
    case BAUDRATE_INVOKE_1MHZ_MSP430:
      BSL_invoke_sequence();
    case BAUDRATE_INVOKE_1MHZ_MSP432E4XX:
    case BAUDRATE_EXIT_1MHZ_MSP432E4XX:
    case BAUDRATE_1MHZ:
      UCB1BRW = 4;
      break;

    default:
      return 1;
      break;
  }
    
  if ((lBaudrate == BAUDRATE_INVOKE_125KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_INVOKE_250KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_INVOKE_500KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_INVOKE_1MHZ_MSP432E4XX))
  {
    isE4xx = 1;
  }
  
  if ((lBaudrate == BAUDRATE_EXIT_125KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_EXIT_250KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_EXIT_500KHZ_MSP432E4XX) || 
      (lBaudrate == BAUDRATE_EXIT_1MHZ_MSP432E4XX)   ||
      (lBaudrate == BAUDRATE_INVOKE_125KHZ_MSP430)   || 
      (lBaudrate == BAUDRATE_INVOKE_250KHZ_MSP430)   || 
      (lBaudrate == BAUDRATE_INVOKE_500KHZ_MSP430)   || 
      (lBaudrate == BAUDRATE_INVOKE_1MHZ_MSP430))
  {
    isE4xx = 0;
  }
  
  if (isE4xx)
  {
    // CKPH=0, idle low (CKPL=0), MSB first, master, 3-pin SPI, synchronous mode
    UCB1CTL0 |= UCMSB + UCMST + UCMODE_0 + UCSYNC;
  }
  else
  {
    // CKPH=0, idle high (CKPL=1), MSB first, master, 3-pin SPI, synchronous mode
    UCB1CTL0 |= UCCKPL + UCMSB + UCMST + UCMODE_0 + UCSYNC;
  } 

    UCB1CTL1 &= ~UCSWRST;                   // Initialize USCI state machine

    return 0;
}

void spiControl()
{
  static uint16_t TXresponseLength = 0;
  static uint16_t TXresponseIndex = 0;
  static uint16_t RXresponseLength = 0;
  static uint16_t RXresponseIndex = 0;
  static enum spiStates spiState = idle;
  uint8_t rxChar = 0xFF;

  switch(spiState)
  {
    case idle:
      // Make sure all transmission are done before 
      // disabling the SPI slave enable
      while(UCB1STAT & UCBUSY); 
      P4OUT |= BIT3;  // Disable SPI slave
      
      // Send data from USB buffer to SPI 
      if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
      {
        P4OUT &= ~BIT3;  // Enable SPI slave
        uint8_t txChar = 0;
        UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
        USBCDC_receiveData(&txChar, 1, CDC0_INTFNUM);
        while(!(UCB1IFG & UCTXIFG));
        UCB1TXBUF = txChar;
        LED2_TOGGLE
          
        if (isE4xx)
        {
          isHostAck = 0;
          UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
          if(txChar == 0x00)  // Check for BSL header from host
          {
            TXresponseLength = 1;
            isHostAck = 1;
          }
          else 
          {
            TXresponseLength = txChar-1;
            isHostAck = 0;                    
          }
          spiState = TXreceive;          
        }
        else
        {
          if(txChar == 0x80)  // Check for BSL header from host
          {
            UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
            spiState = TXlengthLow;
          }
          else
          {
            while(!(UCB1IFG & UCRXIFG));
            // Store received byte in TX buffer
            *txBuffer_write_ptr = UCB1RXBUF;
            txBuffer_write_ptr++;
            if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
            {
                txBuffer_write_ptr = txBuffer;
            }
          }          
        }
      }
      break;

    case TXlengthLow:
      if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
      {
          uint8_t txChar = 0;
          UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
          USBCDC_receiveData(&txChar, 1, CDC0_INTFNUM);
          while(!(UCB1IFG & UCTXIFG));
          UCB1TXBUF = txChar;
          LED2_TOGGLE
          TXresponseLength = (uint16_t)txChar;
          spiState = TXlengthHigh;
      }
      break;

    case TXlengthHigh:
      if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
      {
          uint8_t txChar = 0;
          UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
          USBCDC_receiveData(&txChar, 1, CDC0_INTFNUM);
          while(!(UCB1IFG & UCTXIFG));
          UCB1TXBUF = txChar;
          LED2_TOGGLE
          TXresponseLength |= (uint16_t)txChar << 8;
          spiState = TXreceive;
      }
      break;

    case TXreceive:
      if (isE4xx)
      {
        // Host (PC) sends the ACK (0x00 and 0xCC) to the target
        // without expecting any BSL response
        if (isHostAck)
        {
          // Send data from USB buffer to SPI
          if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
          {
            uint8_t txChar = 0;
            UCB1IFG &= ~UCRXIFG;       // Clear RX IRQ flag
            USBCDC_receiveData(&txChar, 1, CDC0_INTFNUM);
            while(!(UCB1IFG & UCTXIFG));
            UCB1TXBUF = txChar;
            while(UCB1STAT & UCBUSY);  // Wait for completed send operation
            LED2_TOGGLE
            spiState = idle;
            TXresponseLength = 0;
            TXresponseIndex = 0;
          }
        }
        else
        {
          uint8_t tmpBuf[2];
          tmpBuf[0] = 0x00;
          tmpBuf[1] = 0x00;
          
          // Wait for BSL command to be send from host completely
          while(TXresponseIndex < TXresponseLength)  
          {
            /* Send data from USB buffer to SPI */
            if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
            {
                uint8_t txChar = 0;
                UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
                USBCDC_receiveData(&txChar, 1, CDC0_INTFNUM);
                while(!(UCB1IFG & UCTXIFG));
                UCB1TXBUF = txChar;
                if (TXresponseIndex<2)
                {
                  tmpBuf[TXresponseIndex]=txChar;
                }
                LED2_TOGGLE
                TXresponseIndex++;
            }
            
            if (tmpBuf[0]==BSL_CMD_GET_STATUS && tmpBuf[1]==BSL_CMD_GET_STATUS )
            {
              RXresponseLength=BSL_E4xx_RESPONSE_GET_STATUS; 
            }
            else
            {
              RXresponseLength=BSL_E4xx_RESPONSE_ACK;
            }
          }
          
          if(TXresponseIndex==TXresponseLength)
          {
            rxChar = 0xFF;
            while (rxChar == 0xFF)
            {
                UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
                // Shift SPI to receive BSL response
                while(!(UCB1IFG & UCTXIFG));
                UCB1TXBUF = 0x00;
                while(UCB1STAT & UCBUSY);  // Wait for completed send operation
                // Send response to host
                while(!(UCB1IFG & UCRXIFG));
                rxChar = UCB1RXBUF;
                __delay_cycles(SPI_DELAY_CLK_PULSE_E4XX);
            }
            
            // Store received byte in TX buffer
            *txBuffer_write_ptr = rxChar;
            txBuffer_write_ptr++;
            if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
            {
                txBuffer_write_ptr = txBuffer;
            }
            if(rxChar != BSL_RESPONSE_ACK)  // BSL NACK
            {
                spiState = idle;
            }
            else  // BSL ACK Retrieve and send BSL core response
            {
                spiState = RXheader;
                RXresponseLength--;
            }
            TXresponseLength = 0;
            TXresponseIndex = 0;
          }
        }      
      }
      else
      {
        while(TXresponseIndex < TXresponseLength + BSL_CRC_LENGTH)  // Wait for BSL command to be send from host completely
        {
          /* Send data from USB buffer to SPI */
          if(USBCDC_bytesInUSBBuffer(CDC0_INTFNUM) > 0)
          {
              UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
              while(!(UCB1IFG & UCTXIFG));
              USBCDC_receiveData((uint8_t *)&UCB1TXBUF, 1, CDC0_INTFNUM);
              LED2_TOGGLE
              TXresponseIndex++;
          }
        }
        if(TXresponseIndex == TXresponseLength + BSL_CRC_LENGTH)
        {
          uint8_t rxChar = 0xFF;
          
          while (rxChar == 0xFF)
          {
              UCB1IFG &= ~UCRXIFG;  // Clear RX IRQ flag
            
              // Shift SPI to receive BSL response
              while(!(UCB1IFG & UCTXIFG));
              UCB1TXBUF = 0xFF;
              while(UCB1STAT & UCBUSY);  // Wait for completed send operation
              // Send response to host
              while(!(UCB1IFG & UCRXIFG));
              rxChar = UCB1RXBUF;
              __delay_cycles(SPI_DELAY_CLK_PULSE_P4XX);
          }
          
          // Store received byte in TX buffer
          *txBuffer_write_ptr = rxChar;
          txBuffer_write_ptr++;
          if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
          {
              txBuffer_write_ptr = txBuffer;
          }
          if(rxChar != 0x00)  // BSL NACK
          {
              spiState = idle;
          }
          else  // BSL ACK Retrieve and send BSL core response
          {
              spiState = RXheader;
          }
          TXresponseLength = 0;
          TXresponseIndex = 0;
        }      
      }
      break;

    case RXheader:
      if (isE4xx)
      {
        // Shift SPI to receive BSL response
        while(!(UCB1IFG & UCTXIFG));
        UCB1TXBUF = 0x00;
        while(UCB1STAT & UCBUSY);  // Wait for completed send operation
        // Send response to host
        while(!(UCB1IFG & UCRXIFG));
        rxChar = UCB1RXBUF;
        for(uint8_t retryIndex = 50; retryIndex > 0; retryIndex--)
        {
            if(rxChar != BSL_RESPONSE_ACK_E4xx)
            {
               while(rxChar != BSL_RESPONSE_ACK_E4xx)
               {
                  // Shift SPI to receive BSL response
                  while(!(UCB1IFG & UCTXIFG));
                  UCB1TXBUF = 0x00;
                  while(UCB1STAT & UCBUSY);  // Wait for completed send operation
                  // Send response to host
                  while(!(UCB1IFG & UCRXIFG));
                  rxChar = UCB1RXBUF; 
                  __delay_cycles(SPI_DELAY_CLK_PULSE_E4XX);
               }
               RXresponseLength--;
            } 
            else 
            {
                break;
            }
        }
        // Store received byte in TX buffer
        *txBuffer_write_ptr = rxChar;
        txBuffer_write_ptr++;
        
        if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
        {
            txBuffer_write_ptr = txBuffer;
        }
        spiState = RXreceive;         
      }
      else
      {
        // Shift SPI to receive BSL response
        while(!(UCB1IFG & UCTXIFG));
        UCB1TXBUF = 0xFF;
        while(UCB1STAT & UCBUSY);  // Wait for completed send operation
        // Send response to host
        while(!(UCB1IFG & UCRXIFG));
        rxChar = UCB1RXBUF;
        for(uint8_t retryIndex = 50; retryIndex > 0; retryIndex--)
        {
            if(rxChar != 0x80)
            {
               while(rxChar != 0x80)
               {
                  // Shift SPI to receive BSL response
                  while(!(UCB1IFG & UCTXIFG));
                  UCB1TXBUF = 0xFF;
                  while(UCB1STAT & UCBUSY);  // Wait for completed send operation
                  // Send response to host
                  while(!(UCB1IFG & UCRXIFG));
                  rxChar = UCB1RXBUF; 
                  __delay_cycles(SPI_DELAY_CLK_PULSE_P4XX);
               }
            } 
            else 
            {
                break;
            }
        }
        // Store received byte in TX buffer
        *txBuffer_write_ptr = rxChar;
        txBuffer_write_ptr++;
        if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
        {
            txBuffer_write_ptr = txBuffer;
        }
        spiState = RXlengthLow;      
      }
      break;

    case RXlengthLow:
      // Shift SPI to receive BSL response
      while(!(UCB1IFG & UCTXIFG));
      UCB1TXBUF = 0xFF;
      while(UCB1STAT & UCBUSY);  // Wait for completed send operation
      // Send response to host
      while(!(UCB1IFG & UCRXIFG));
      rxChar = UCB1RXBUF;
      RXresponseLength = (uint16_t)rxChar;
      // Store received byte in TX buffer
      *txBuffer_write_ptr = rxChar;
      txBuffer_write_ptr++;
      if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
      {
          txBuffer_write_ptr = txBuffer;
      }
      spiState = RXlengthHigh;
      break;

    case RXlengthHigh:
      // Shift SPI to receive BSL response
      while(!(UCB1IFG & UCTXIFG));
      UCB1TXBUF = 0xFF;
      while(UCB1STAT & UCBUSY);  // Wait for completed send operation
      // Send response to host
      while(!(UCB1IFG & UCRXIFG));
      rxChar = UCB1RXBUF;
      RXresponseLength |= (uint16_t)rxChar << 8;
      // Store received byte in TX buffer
      *txBuffer_write_ptr = rxChar;
      txBuffer_write_ptr++;
      if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
      {
          txBuffer_write_ptr = txBuffer;
      }
      spiState = RXreceive;
      break;

    case RXreceive:
      if (isE4xx)
      {
        if(RXresponseIndex < RXresponseLength)  // Wait for BSL response to be received completely
        {
          // Shift SPI to receive BSL response
          while(!(UCB1IFG & UCTXIFG));
          UCB1TXBUF = 0x00;
          while(UCB1STAT & UCBUSY);  // Wait for completed send operation
          // Send response to host
          while(!(UCB1IFG & UCRXIFG));
          // Store received byte in TX buffer
          *txBuffer_write_ptr = UCB1RXBUF;
          txBuffer_write_ptr++;
          if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
          {
              txBuffer_write_ptr = txBuffer;
          }
          RXresponseIndex++;
        }
        
        if(RXresponseIndex == RXresponseLength)
        {
          RXresponseLength = 0;
          RXresponseIndex = 0;
          spiState = idle;
        }  
      }
      else
      {
        if(RXresponseIndex < RXresponseLength + BSL_CRC_LENGTH)  // Wait for BSL response to be received completely
        {
          // Shift SPI to receive BSL response
          while(!(UCB1IFG & UCTXIFG));
          UCB1TXBUF = 0xFF;
          while(UCB1STAT & UCBUSY);  // Wait for completed send operation
          // Send response to host
          while(!(UCB1IFG & UCRXIFG));
          // Store received byte in TX buffer
          *txBuffer_write_ptr = UCB1RXBUF;
          txBuffer_write_ptr++;
          if(txBuffer_write_ptr - txBuffer == TX_BUFFER_SIZE)
          {
              txBuffer_write_ptr = txBuffer;
          }
          RXresponseIndex++;
        }
        if(RXresponseIndex == RXresponseLength + BSL_CRC_LENGTH)
        {
          RXresponseLength = 0;
          RXresponseIndex = 0;
          spiState = idle;
        }
      }
        break;

      default:
        break;
    }

  /* Send data from USB buffer to host */
  if(txBuffer_read_ptr - txBuffer == TX_BUFFER_SIZE)
      {
          txBuffer_read_ptr = txBuffer;
      }
  if(txBuffer_read_ptr != txBuffer_write_ptr)
  {
    /* New data in txBuffer, send it */
    uint16_t bytes_to_send = 0;
    // In case the write ptr < the read ptr sent till end of buffer...
    if(txBuffer_write_ptr < txBuffer_read_ptr)
    {
      bytes_to_send = TX_BUFFER_SIZE - (txBuffer_read_ptr - txBuffer);
      if(kUSBCDC_sendStarted == USBCDC_sendData(txBuffer_read_ptr, bytes_to_send, CDC0_INTFNUM))
      {
          txBuffer_read_ptr += bytes_to_send;
      }
    }
    // ... else sent till catching up with the write pts.
    else
    {
      bytes_to_send = txBuffer_write_ptr - txBuffer_read_ptr;
      if(kUSBCDC_sendStarted == USBCDC_sendData(txBuffer_read_ptr, bytes_to_send, CDC0_INTFNUM))
      {
          txBuffer_read_ptr += bytes_to_send;
      }
    }
  }
}
